function callsites() {
  const _prepareStackTrace = Error.prepareStackTrace;
  try {
    let result = [];
    Error.prepareStackTrace = (_, callSites) => {
      const callSitesWithoutCurrent = callSites.slice(1);
      result = callSitesWithoutCurrent;
      return callSitesWithoutCurrent;
    };
    new Error().stack;
    return result;
  } finally {
    Error.prepareStackTrace = _prepareStackTrace;
  }
}
const list = [
  // Native ES errors https://262.ecma-international.org/12.0/#sec-well-known-intrinsic-objects
  Error,
  EvalError,
  RangeError,
  ReferenceError,
  SyntaxError,
  TypeError,
  URIError,
  AggregateError,
  // Built-in errors
  globalThis.DOMException,
  // Node-specific errors
  // https://nodejs.org/api/errors.html
  globalThis.AssertionError,
  globalThis.SystemError
].filter(Boolean).map(
  (constructor) => [constructor.name, constructor]
);
const errorConstructors = new Map(list);
class NonError extends Error {
  name = "NonError";
  constructor(message) {
    super(NonError._prepareSuperMessage(message));
  }
  static _prepareSuperMessage(message) {
    try {
      return JSON.stringify(message);
    } catch {
      return String(message);
    }
  }
}
const errorProperties = [
  {
    property: "name",
    enumerable: false
  },
  {
    property: "message",
    enumerable: false
  },
  {
    property: "stack",
    enumerable: false
  },
  {
    property: "code",
    enumerable: true
  },
  {
    property: "cause",
    enumerable: false
  },
  {
    property: "errors",
    enumerable: false
  }
];
const toJsonWasCalled = /* @__PURE__ */ new WeakSet();
const toJSON = (from) => {
  toJsonWasCalled.add(from);
  const json = from.toJSON();
  toJsonWasCalled.delete(from);
  return json;
};
const newError = (name) => {
  const ErrorConstructor = errorConstructors.get(name) ?? Error;
  return ErrorConstructor === AggregateError ? new ErrorConstructor([]) : new ErrorConstructor();
};
const destroyCircular = ({
  from,
  seen,
  to,
  forceEnumerable,
  maxDepth,
  depth,
  useToJSON,
  serialize
}) => {
  if (!to) {
    if (Array.isArray(from)) {
      to = [];
    } else if (!serialize && isErrorLike(from)) {
      to = newError(from.name);
    } else {
      to = {};
    }
  }
  seen.push(from);
  if (depth >= maxDepth) {
    return to;
  }
  if (useToJSON && typeof from.toJSON === "function" && !toJsonWasCalled.has(from)) {
    return toJSON(from);
  }
  const continueDestroyCircular = (value) => destroyCircular({
    from: value,
    seen: [...seen],
    forceEnumerable,
    maxDepth,
    depth,
    useToJSON,
    serialize
  });
  for (const [key, value] of Object.entries(from)) {
    if (value && value instanceof Uint8Array && value.constructor.name === "Buffer") {
      to[key] = "[object Buffer]";
      continue;
    }
    if (value !== null && typeof value === "object" && typeof value.pipe === "function") {
      to[key] = "[object Stream]";
      continue;
    }
    if (typeof value === "function") {
      continue;
    }
    if (!value || typeof value !== "object") {
      try {
        to[key] = value;
      } catch {
      }
      continue;
    }
    if (!seen.includes(from[key])) {
      depth++;
      to[key] = continueDestroyCircular(from[key]);
      continue;
    }
    to[key] = "[Circular]";
  }
  if (serialize || to instanceof Error) {
    for (const { property, enumerable } of errorProperties) {
      if (from[property] !== void 0 && from[property] !== null) {
        Object.defineProperty(to, property, {
          value: isErrorLike(from[property]) || Array.isArray(from[property]) ? continueDestroyCircular(from[property]) : from[property],
          enumerable: forceEnumerable ? true : enumerable,
          configurable: true,
          writable: true
        });
      }
    }
  }
  return to;
};
function serializeError(value, options = {}) {
  const {
    maxDepth = Number.POSITIVE_INFINITY,
    useToJSON = true
  } = options;
  if (typeof value === "object" && value !== null) {
    return destroyCircular({
      from: value,
      seen: [],
      forceEnumerable: true,
      maxDepth,
      depth: 0,
      useToJSON,
      serialize: true
    });
  }
  if (typeof value === "function") {
    return `[Function: ${value.name || "anonymous"}]`;
  }
  return value;
}
function deserializeError(value, options = {}) {
  const { maxDepth = Number.POSITIVE_INFINITY } = options;
  if (value instanceof Error) {
    return value;
  }
  if (isMinimumViableSerializedError(value)) {
    return destroyCircular({
      from: value,
      seen: [],
      to: newError(value.name),
      maxDepth,
      depth: 0,
      serialize: false
    });
  }
  return new NonError(value);
}
function isErrorLike(value) {
  return Boolean(value) && typeof value === "object" && typeof value.name === "string" && typeof value.message === "string" && typeof value.stack === "string";
}
function isMinimumViableSerializedError(value) {
  return Boolean(value) && typeof value === "object" && typeof value.message === "string" && !Array.isArray(value);
}
var InvokeReturnType = /* @__PURE__ */ ((InvokeReturnType2) => {
  InvokeReturnType2["error"] = "error";
  InvokeReturnType2["result"] = "result";
  return InvokeReturnType2;
})(InvokeReturnType || {});
const channelGenerator = (controller, event, type) => `$electron-ipc-flow$||${type}||${controller}||${event}`;
const ErrorHandler = {
  serialize: serializeError,
  deserialize: deserializeError
};
function assertIsNull(message, value) {
  if (typeof value === "undefined" || value === null) {
    throw new TypeError(message);
  }
}
const isDebug = () => (typeof process === "undefined" ? window : process.env).ELECTRON_IPC_FLOW_DEBUG === "true";
function debug(action, detail, event, channelType) {
  if (!isDebug()) {
    return;
  }
  const firstCallStack = callsites()[1];
  if (typeof firstCallStack === "undefined") {
    return;
  }
  const typeName = firstCallStack.getTypeName();
  const methodName = firstCallStack.getMethodName() ?? firstCallStack.getFunctionName() ?? "<anonymous>";
  const controller = this == null ? void 0 : this.name;
  const obj = {
    controller,
    method: `${typeName ? typeName.concat(".") : ""}${methodName}`,
    action,
    event,
    channel: channelType ? channelGenerator(controller, event ?? "x", channelType) : void 0,
    [detail instanceof Error ? "error" : "detail"]: detail,
    source: `${firstCallStack.getFileName()}:${firstCallStack.getLineNumber()}:${firstCallStack.getColumnNumber()}`
  };
  for (const key of Object.keys(obj)) {
    if (typeof Reflect.get(obj, key) === "undefined") {
      Reflect.deleteProperty(obj, key);
    }
  }
  console.debug(obj);
}
const IpcClientControllerRegistered = /* @__PURE__ */ new Set();
class IpcClientController {
  #name;
  #debug = debug.bind(this);
  #ipcRendererEventListeners = /* @__PURE__ */ new Map();
  #eventsListeners = /* @__PURE__ */ new Map();
  /**
   * @param name {string} Controller name
   */
  constructor(name) {
    var _a;
    if (name.trim() === "") {
      throw new SyntaxError('IpcClientController: "name" cannot be an empty string.');
    }
    this.#name = name;
    const g = typeof globalThis === "undefined" ? window : globalThis;
    if (typeof ((_a = g.$IpcClientController) == null ? void 0 : _a.register) === "function") {
      this.#debug("auto register");
      g.$IpcClientController.register(this.name);
    }
  }
  /**
   * Controller name
   */
  get name() {
    return this.#name;
  }
  #addEventListener(event, listener, once = false) {
    if (!this.#ipcRendererEventListeners.has(event)) {
      this.#debug("add global event listener", null, event, "c");
      const globalListener = this.#ipcRendererEventListener.bind(this, event);
      IpcClientController.#getGlobalIpcController().on(this.name, event, globalListener);
      this.#ipcRendererEventListeners.set(event, globalListener);
    }
    this.#debug("add event listener", { once }, event, "c");
    const listeners = [
      ...this.#eventsListeners.get(event) ?? [],
      {
        listener,
        once
      }
    ];
    this.#eventsListeners.set(event, listeners);
  }
  #ipcRendererEventListener(event, eventObj, ...args) {
    this.#debug("received", { args }, event, "c");
    const listeners = this.#eventsListeners.get(event) ?? [];
    const onceListeners = listeners.filter((item) => {
      try {
        item.listener(eventObj, ...args);
      } catch (error) {
        this.#debug("catch error", error, event, "c");
      }
      return item.once;
    });
    for (const item of onceListeners) {
      this.off(event, item.listener);
    }
  }
  /**
   * Use `ipcRenderer.invoke()` to call the function defined by `IpcServerController.handle()`.
   */
  async invoke(name, ...args) {
    this.#debug("invoke", { args }, name, "i");
    const result = await IpcClientController.#getGlobalIpcController().invoke(this.name, name, ...args);
    this.#debug("received", result, name, "i");
    if (result.type === InvokeReturnType.error) {
      throw ErrorHandler.deserialize(result.value);
    }
    return result.value;
  }
  postMessage(event, message, transfer) {
    this.#debug("postMessage", { message }, event, "s");
    globalThis.postMessage(
      {
        $action: "postMessage.send",
        controller: this.name,
        event,
        message
      },
      "*",
      transfer
    );
  }
  /**
   * Removes event listeners added using `on()` and `once()`.
   *
   * If the `listener` parameter is not provided, all listeners for
   * the corresponding event will be removed.
   */
  off(event, listener) {
    if (!this.#ipcRendererEventListeners.has(event)) {
      return;
    }
    this.#debug("remove event listener", null, event, "c");
    const listeners = (listener ? this.#eventsListeners.get(event) ?? [] : []).filter(
      (item) => item.listener !== listener
    );
    this.#eventsListeners.set(event, listeners);
  }
  /**
   * Adds a specific client event listener, where the first parameter of
   * the listener is of type `Electron.IpcRendererEvent`.
   */
  on(event, listener) {
    this.#addEventListener(event, listener);
  }
  /**
   * Like `on()`, but the listener is removed after it is triggered once.
   *
   * @see {on}
   */
  once(event, listener) {
    this.#addEventListener(event, listener, true);
  }
  /**
   * Uses `ipcRenderer.send()` to send to the server event listeners added with `IpcServerController.on()`.
   */
  send(event, ...args) {
    this.#debug("send", { args }, event, "s");
    IpcClientController.#getGlobalIpcController().send(this.name, event, ...args);
  }
  /**
   * If auto register is not enabled, use this method in the
   * **preload script** to manually register the controller.
   */
  register() {
    IpcClientControllerRegistered.add(this.name);
  }
  /**
   * Can use this method in the **preload script** to unregister the controller
   */
  unregister() {
    IpcClientControllerRegistered.delete(this.name);
  }
  static #getGlobalIpcController() {
    const global = typeof window !== "undefined" ? window : globalThis;
    assertIsNull(
      `IpcClientController: Can't find the "globalThis.$IpcClientController", Forgot to use "preloadInit" in the preload script?`,
      global.$IpcClientController
    );
    return global.$IpcClientController;
  }
}
function createIpcClient(...args) {
  const controller = new IpcClientController(...args);
  return new Proxy(controller, {
    get(target, p) {
      if (typeof p === "string" && p.startsWith("$")) {
        return target.invoke.bind(target, p.substring(1));
      }
      const result = Reflect.get(target, p);
      return typeof result === "function" ? result.bind(target) : result;
    }
  });
}
function preloadInit$1(ipcRenderer, autoRegister) {
  function checkControllerRegistered(controllerName) {
    if (!IpcClientControllerRegistered.has(controllerName)) {
      throw new RangeError(`IpcClientController "${controllerName}" not registered.`);
    }
  }
  const obj = {
    invoke(controllerName, name, ...args) {
      checkControllerRegistered(controllerName);
      return ipcRenderer.invoke(channelGenerator(controllerName, name, "i"), ...args);
    },
    on(controllerName, event, listener) {
      checkControllerRegistered(controllerName);
      ipcRenderer.on(channelGenerator(controllerName, event, "c"), listener);
    },
    register(name) {
      IpcClientControllerRegistered.add(name);
    },
    send(controllerName, event, ...args) {
      checkControllerRegistered(controllerName);
      ipcRenderer.send(channelGenerator(controllerName, event, "s"), ...args);
    }
  };
  if (!autoRegister) {
    Reflect.deleteProperty(obj, "register");
  }
  return {
    api: obj,
    key: "$IpcClientController"
  };
}
function isPromise(obj) {
  return !!obj && (typeof obj === "object" || typeof obj === "function") && typeof obj.then === "function";
}
class IpcServerController {
  /**
   * `IpcServerController` calls the trust handler when it receives a call or event.
   * If the trust handler returns false, an exception is thrown to the renderer process: "*Blocked by trust handler.*".
   *
   *
   * @see {trustHandler}
   */
  static TrustHandler = () => true;
  /**
   * Used to specify which renderers the `IpcServerController.send` method sends events to.
   * by default, it does not send to any renderers.
   *
   * @see {webContentsGetter}
   */
  static WebContentsGetter;
  /**
   * `Electron.IpcMain` used by all IpcServerController instances.
   *
   * @see {Electron.IpcMain}
   */
  static IpcMain;
  /**
   * `Electron.IpcMain` used by this instance.
   *
   * @see {Electron.IpcMain}
   * @see {IpcServerController.IpcMain}
   */
  ipcMain;
  /**
   * This controller specific trust handler.
   *
   * @see {TrustHandler}
   */
  trustHandler;
  /**
   * This controller specific webContents getter.
   *
   * @see {WebContentsGetter}
   */
  webContentsGetter;
  #name;
  #debug = debug.bind(this);
  #ipcMainEventListeners = /* @__PURE__ */ new Map();
  #eventsListeners = /* @__PURE__ */ new Map();
  #handlers;
  /**
   * @param name {string} Controller name
   */
  constructor(name) {
    if (name.trim() === "") {
      throw new SyntaxError('IpcServerController: "name" cannot be an empty string.');
    }
    this.#name = name;
  }
  /**
   * Controller name
   */
  get name() {
    return this.#name;
  }
  /**
   * Use the `handle()` to set all properties of an object as function handlers, with the
   * property names being used as function names.
   *
   * If the value is set to `undefined` or if the property has been set previously, the
   * existing function handler will be removed using `removeHandler()` first.
   *
   * @see {handle}
   * @see {removeHandler}
   */
  set handlers(value) {
    const getHandlersKeys = (handlers) => typeof handlers === "object" ? /* @__PURE__ */ new Set([
      ...Object.getOwnPropertyNames(handlers).filter((v) => v !== "constructor"),
      ...Object.getOwnPropertyNames(Object.getPrototypeOf(handlers)).filter((v) => v !== "constructor")
    ]) : /* @__PURE__ */ new Set();
    if (typeof this.#handlers === "object") {
      for (const name of getHandlersKeys(this.#handlers)) {
        if (typeof this.#handlers[name] === "function") {
          this.removeHandler(name);
        }
      }
    }
    this.#handlers = value;
    if (typeof this.#handlers === "object") {
      for (const name of getHandlersKeys(this.#handlers)) {
        if (typeof name === "string" && typeof this.#handlers[name] === "function") {
          this.handle(name, this.#handlers[name].bind(this.#handlers));
        }
      }
    }
  }
  get #ipc() {
    const value = this.ipcMain ?? IpcServerController.IpcMain;
    if (value === void 0) {
      throw new TypeError("IpcServerController: static.IpcMain or instance.ipcMain must be set");
    }
    return value;
  }
  get #trustHandler() {
    return this.trustHandler ?? IpcServerController.TrustHandler;
  }
  #clientEventChannel(event) {
    return channelGenerator(this.name, event, "c");
  }
  #invokeChannel(name) {
    return channelGenerator(this.name, name, "i");
  }
  #serverEventChannel(event) {
    return channelGenerator(this.name, event, "s");
  }
  #addEventListener(event, listener, once = false) {
    const channel = this.#serverEventChannel(event);
    if (!this.#ipcMainEventListeners.has(event)) {
      this.#debug("add global event listener", null, event, "s");
      const globalListener = this.#ipcMainEventListener.bind(this, event);
      this.#ipc.on(channel, globalListener);
      this.#ipcMainEventListeners.set(event, globalListener);
    }
    this.#debug("add event listener", { once }, event, "s");
    const listeners = [
      ...this.#eventsListeners.get(event) ?? [],
      {
        listener,
        once
      }
    ];
    this.#eventsListeners.set(event, listeners);
  }
  async #ipcMainEventListener(event, eventObj, ...args) {
    try {
      const trust = this.#trustHandler(this, event, "event", eventObj);
      if (trust === false || !await trust) {
        this.#debug("blocked", null, event, "s");
        return;
      }
    } catch (err) {
      console.error("An error occurred in the trust handler:", err);
      return;
    }
    this.#debug("received", { args }, event, "s");
    const listeners = this.#eventsListeners.get(event) ?? [];
    const onceListeners = listeners.filter((item) => {
      try {
        item.listener(eventObj, ...args);
      } catch (error) {
        this.#debug("catch error", error, event, "s");
      }
      return item.once;
    });
    for (const item of onceListeners) {
      this.off(event, item.listener);
    }
  }
  /**
   * Why use a callback method? If getter is not a Promise, it can be call sync.
   */
  #callWebContentsGetter(cb) {
    try {
      const getter = this.webContentsGetter ?? IpcServerController.WebContentsGetter ?? (() => []);
      const result = getter();
      if (isPromise(result)) {
        result.then(cb).catch((err) => {
          console.error("An error occurred in the webContents getter:", err);
        });
      } else {
        cb(result);
      }
    } catch (err) {
      console.error("An error occurred in the webContents getter:", err);
    }
  }
  async #handle(name, passEvent, func, event, ...args) {
    this.#debug("received", { args }, name, "i");
    try {
      const trust = this.#trustHandler(this, name, "invoke", event);
      if (trust === false || !await trust) {
        this.#debug("blocked", null, name, "i");
        return {
          type: InvokeReturnType.error,
          value: ErrorHandler.serialize(new Error("Blocked by trust handler"))
        };
      }
    } catch (err) {
      console.error("An error occurred in the trust handler:", err);
      return {
        type: InvokeReturnType.error,
        value: ErrorHandler.serialize(new Error("Blocked by trust handler"))
      };
    }
    try {
      let value = passEvent ? func(event, ...args) : func(...args);
      if (isPromise(value)) {
        value = await value;
      }
      this.#debug("send result", { value }, name, "i");
      return {
        type: InvokeReturnType.result,
        value
      };
    } catch (err) {
      this.#debug("catch error", err, name, "i");
      return {
        type: InvokeReturnType.error,
        value: ErrorHandler.serialize(err)
      };
    }
  }
  #send(frameId, event, ...args) {
    this.#callWebContentsGetter((items) => {
      const channel = this.#clientEventChannel(event);
      for (const webContents of items) {
        this.#debug("send", { args, webContents: webContents.id, frameId }, event, "c");
        try {
          if (frameId === null) {
            webContents.send(channel, ...args);
          } else {
            webContents.sendToFrame(frameId, channel, ...args);
          }
        } catch (err) {
          console.error(`An error occurred in the webContents.${frameId === null ? "send" : "sendToFrame"}:`, err);
        }
      }
    });
  }
  /**
   * Uses `IpcMain.handle()` to add a specific function handler.
   */
  handle(name, handler) {
    this.#debug("add function handler", null, name, "i");
    this.#ipc.handle(this.#invokeChannel(name), this.#handle.bind(this, name, false, handler));
  }
  /**
   * Like `handle()`, it will just pass the event object.
   *
   * @see {handle}
   */
  handleWithEvent(name, handler) {
    this.#debug("add function handler with event", null, name, "i");
    this.#ipc.handle(this.#invokeChannel(name), this.#handle.bind(this, name, true, handler));
  }
  /**
   * Uses `IpcMain.removeHandler()` to remove a specific function handler.
   */
  removeHandler(name) {
    this.#debug("remove function handler", null, name, "i");
    this.#ipc.removeHandler(this.#invokeChannel(name));
  }
  /**
   * Removes event listeners added using `on()` and `once()`.
   *
   * If the `listener` parameter is not provided, all listeners for
   * the corresponding event will be removed.
   */
  off(event, listener) {
    if (!this.#ipcMainEventListeners.has(event)) {
      return;
    }
    const channel = this.#serverEventChannel(event);
    this.#debug("remove event listener", null, event, "s");
    const handlers = (listener ? this.#eventsListeners.get(event) ?? [] : []).filter(
      (item) => item.listener !== listener
    );
    if (handlers.length === 0) {
      const listener2 = this.#ipcMainEventListeners.get(event);
      if (listener2) {
        this.#debug("remove global event listener", null, event, "s");
        this.#ipc.removeListener(channel, listener2);
        this.#ipcMainEventListeners.delete(event);
      }
      this.#eventsListeners.delete(event);
    } else {
      this.#eventsListeners.set(event, handlers);
    }
  }
  /**
   * Adds a specific server event listener, where the first parameter of
   * the listener is of type `Electron.IpcMainEvent`.
   */
  on(event, listener) {
    this.#addEventListener(event, listener);
  }
  /**
   * Like `on()`, but the listener is removed after it is triggered once.
   *
   * @see {on}
   */
  once(event, listener) {
    this.#addEventListener(event, listener, true);
  }
  /**
   * Uses `WebContents.send()` to send to the server event listeners added with `IpcClientController.on()`.
   */
  send(event, ...args) {
    this.#send(null, event, ...args);
  }
  /**
   * Like `send()`, but use `WebContents.sendToFrame()`.
   *
   * @see {send}
   */
  sendToFrame(frameId, event, ...args) {
    this.#send(frameId, event, ...args);
  }
}
function createIpcServer(...args) {
  return new IpcServerController(...args);
}
function preloadInit(contextBridge, ipcRenderer, options) {
  const opt = {
    autoRegister: true,
    isolatedWorldId: 0,
    ...options ?? {}
  };
  const items = [
    {
      api: isDebug() ? "true" : "false",
      key: "ELECTRON_IPC_FLOW_DEBUG"
    },
    preloadInit$1(ipcRenderer, opt.autoRegister)
  ];
  for (const item of items) {
    if (opt.isolatedWorldId > 0) {
      contextBridge.exposeInIsolatedWorld(opt.isolatedWorldId, item.key, item.api);
    } else {
      contextBridge.exposeInMainWorld(item.key, item.api);
    }
  }
}
export {
  ErrorHandler,
  IpcClientController,
  IpcServerController,
  channelGenerator,
  createIpcClient,
  createIpcServer,
  preloadInit
};
