"use strict";
const renderer = require("electron/renderer");
const channelGenerator = (controller, event, type) => `$electron-ipc-flow$||${type}||${controller}||${event}`;
const isDebug = () => (typeof process === "undefined" ? window : process.env).ELECTRON_IPC_FLOW_DEBUG === "true";
const IpcClientControllerRegistered = /* @__PURE__ */ new Set();
function preloadInit$1(ipcRenderer, autoRegister) {
  function checkControllerRegistered(controllerName) {
    if (!IpcClientControllerRegistered.has(controllerName)) {
      throw new RangeError(`IpcClientController "${controllerName}" not registered.`);
    }
  }
  const obj = {
    invoke(controllerName, name, ...args) {
      checkControllerRegistered(controllerName);
      return ipcRenderer.invoke(channelGenerator(controllerName, name, "i"), ...args);
    },
    on(controllerName, event, listener) {
      checkControllerRegistered(controllerName);
      ipcRenderer.on(channelGenerator(controllerName, event, "c"), listener);
    },
    register(name) {
      IpcClientControllerRegistered.add(name);
    },
    send(controllerName, event, ...args) {
      checkControllerRegistered(controllerName);
      ipcRenderer.send(channelGenerator(controllerName, event, "s"), ...args);
    }
  };
  if (!autoRegister) {
    Reflect.deleteProperty(obj, "register");
  }
  return {
    api: obj,
    key: "$IpcClientController"
  };
}
function preloadInit(contextBridge, ipcRenderer, options) {
  const opt = {
    autoRegister: true,
    isolatedWorldId: 0,
    ...options ?? {}
  };
  const items = [
    {
      api: isDebug() ? "true" : "false",
      key: "ELECTRON_IPC_FLOW_DEBUG"
    },
    preloadInit$1(ipcRenderer, opt.autoRegister)
  ];
  for (const item of items) {
    if (opt.isolatedWorldId > 0) {
      contextBridge.exposeInIsolatedWorld(opt.isolatedWorldId, item.key, item.api);
    } else {
      contextBridge.exposeInMainWorld(item.key, item.api);
    }
  }
}
preloadInit(renderer.contextBridge, renderer.ipcRenderer, {
  autoRegister: true
});
